from optparse import make_option
from django.core.management.base import BaseCommand
from django.db import transaction
from django.contrib.auth.models import User
from avocado.conf import settings
from avocado.models import DataContext, DataView


class Command(BaseCommand):
    """
    SYNOPSIS::

        python manage.py avocado history [options]

    DESCRIPTION:

        Utilities for managing the history API.

    OPTIONS:

        ``--prune`` - Prunes archived DataContext and DataView entries
        that exceeds HISTORY_MAX_SIZE.

    """

    help = 'Utilities for managing the history API'

    option_list = BaseCommand.option_list + (
        make_option('--prune', action='store_true',
            help='Prunes archived DataContext and DataView entries '
                'that exceeds HISTORY_MAX_SIZE'),
    )

    @transaction.commit_on_success
    def handle(self, *args, **options):
        prune = options.get('prune')

        if not prune:
            print('No option specified, nothing to do')
            return

        if not settings.HISTORY_ENABLED:
            print('History not enabled, nothing to do.')
            return

        if not settings.HISTORY_MAX_SIZE:
            print('Unlimited history is allowed, nothing to do.')
            return

        # Base queryset of archived objects ordered by newest ones first
        views = DataView.objects.filter(archived=True).order_by('-modified')
        contexts = DataContext.objects.filter(archived=True).order_by('-modified')

        max_size = settings.HISTORY_MAX_SIZE

        # Note, this nested call is due to not being able to call delete()
        # when using a limit or offset, thus we must nest the query based on
        # primary key.
        delete = lambda x, y: x.objects.filter(pk__in=y).delete()

        # Prune archived objects from anonymous sessions
        delete(DataView, views.filter(user=None)[max_size:])
        delete(DataContext, contexts.filter(user=None)[max_size:])

        # Prune archived objects per user
        for user in User.objects.all():
            delete(DataView, views.filter(user=user)[max_size:])
            delete(DataContext, contexts.filter(user=user)[max_size:])



from django.template import Context
from django.template.loader import get_template
from _base import BaseExporter


class HTMLExporter(BaseExporter):
    file_extension = 'html'
    content_type = 'text/html'
    preferred_formats = ('html', 'string')

    def write(self, iterable, buff=None, template=None):
        buff = self.get_file_obj(buff)
        generator = self.read(iterable)

        if template:
            context = Context({'rows': generator})
            if isinstance(template, basestring):
                template = get_template(template)
            buff.write(template.render(context))
        else:
            for row in generator:
                for item in row:
                    buff.write(item)
        return buff



import csv
from _base import BaseExporter


class CSVExporter(BaseExporter):
    file_extension = 'csv'
    content_type = 'text/csv'
    preferred_formats = ('number', 'string')

    def write(self, iterable, buff=None):
        header = []
        buff = self.get_file_obj(buff)
        writer = csv.writer(buff, quoting=csv.QUOTE_MINIMAL)

        for i, row_gen in enumerate(self.read(iterable)):
            row = []
            for data in row_gen:
                if i == 0:
                    header.extend(data.keys())
                row.extend(data.values())
            if i == 0:
                writer.writerow(header)
            writer.writerow(row)
        return buff



from optparse import make_option
from django.core.management.base import NoArgsCommand
from django.db import transaction
from avocado.models import DataField


class Command(NoArgsCommand):
    """
    SYNOPSIS::

        python manage.py avocado orphaned [options...]

    DESCRIPTION:

        Determines if any ``Fields`` are no longer valid. This is
        typically only necessary when a data model change occurs. An orphaned
        ``DataField`` will be marked if they are currently published
        (``[P]``).

    OPTIONS:

        ``--unpublish`` - unpublishes orphaned ``DataField`` if currently
        published

    """

    help = "Determines outdated or orphaned Fields."

    option_list = NoArgsCommand.option_list + (
        make_option('--unpublish', action='store_true',
            dest='unpublish', default=False,
            help='Unpublishes orphaned Fields'),
    )

    def _print(self, fields, msg, unpublish=False):
        print
        print '{0}:\n'.format(msg)
        for f in fields:
            print '\t',
            if f.published:
                if unpublish:
                    f.published = False
                    f.save()
                print '[P]',
            else:
                print '   ',
            print f
        print

    @transaction.commit_on_success
    def handle_noargs(self, **options):
        unpublish = options.get('unpublish')
        verbosity = options.get('verbosity')

        unknown_models = []
        unknown_fields = []

        for f in DataField.objects.iterator():
            if f.model is None:
                unknown_models.append(f)
            elif f.field is None:
                unknown_fields.append(f)

        if verbosity:
            if not unknown_models and not unknown_fields:
                print '0 fields orphaned'
            else:
                print '\nKey:\n'
                print '\t[P] - field is published\n'
                if unknown_models:
                    self._print(unknown_models, 'The following Fields have an unknown model', unpublish)
                if unknown_fields:
                    self._print(unknown_fields, 'The following Fields have an unknown field', unpublish)



import re
from django.db.models import get_model
from django.core.management.base import BaseCommand, CommandError
from avocado.models import DataField
from avocado.lexicon.managers import coerce_float

non_alnum = re.compile(r'[^a-zA-Z0-9\-]+')


class Command(BaseCommand):
    """
    SYNOPSIS::

        python manage.py avocado lexicon [options...] field model

    DESCRIPTION:

        Creates a lexicon derived from `field` into the Lexicon
        subclass `model`. The distinct list of values for `field`
        will be queried and loaded into the table.
    """

    help = '\n'.join([
        'Creates a lexicon derived from `field` into the Lexicon',
        'subclass `model`. The distinct list of values for `field`',
        'will be queried and loaded into the table.'
    ])

    args = 'field model'

    def handle(self, field_label, model_label, **options):
        model = get_model(*model_label.split('.'))

        if model is None:
            raise CommandError('Not model named {0} was found'.format(model_label))

        toks = field_label.split('.')
        if len(toks) != 3:
            raise CommandError('{0} is not a valid field identifier. Use a ' \
                '"." delimited notation, e.g. "app.model.field"'.format(field_label))
        datafield = DataField(app_name=toks[0], model_name=toks[1],
            field_name=toks[2])

        if not datafield.field:
            raise CommandError('The field {0} could not be found.'.format(field_label))

        count = 0
        values = list(datafield.values)
        values.sort(key=coerce_float)

        for value in values:
            if value is None or value == '':
                continue
            label = non_alnum.sub(value, ' ').title()
            obj = model(label=label, value=value, order=count, code=count)
            obj.save()
            count += 1

        print '{0} distinct values loaded'.format(count)



from django.core.exceptions import ImproperlyConfigured
from avocado.conf import OPTIONAL_DEPS

if not OPTIONAL_DEPS['scipy']:
    raise ImproperlyConfigured('Numpy and SciPy must be installed to use the clustering utilities.')

import math
import numpy
from scipy.cluster import vq


def find_outliers(wobs, outlier_threshold=3, whitened=True):
    """Finds outliers in the population given some distance threshold.
    Returns the indexes of the outliers from the original array.
    """
    # "Whiten" the observations, e.g. normalize
    if not whitened:
        wobs = vq.whiten(numpy.array(wobs))

    # Get the mid-point of the entire population as the starting centroid
    mid = (len(wobs) - 1) / 2
    center = numpy.sort(wobs, axis=0)[mid:mid + 1]

    # Determine distances from calculated centroid. The outliers are
    # determined by calculating the mean distance across the population
    # and removing the observations that exceed the threshold
    _, dists = vq.vq(wobs, vq.kmeans(wobs, center)[0])
    mean_dist = dists.mean()

    return [i for i, dist in enumerate(dists) \
            if (dist / mean_dist) >= outlier_threshold]


def kmeans_optm(obs, k=None, outlier_threshold=3):
    """Performs k-means clustering on an array of observations. Returns
    the `centroids`, an ordered array of `indexes` corresponding to each
    observations centroid, an ordered array of `distances` of each
    observeration relative to it's centroid, and (optionally) an array of
    outlier indexes (see `find_outliers` above).

    If `k` is not specified, it will be calculated given:

        k = sqrt( n / 2 )

    The initialization method is optimized to find the mid-points of each
    cluster from a sorted array of observations. This ultimately reduces
    the number of iterations required during clustering due to increased
    likelihood of convergence.

    References:
    Number of clusters - http://en.wikipedia.org/wiki/Determining_the_number_of_clusters_in_a_data_set#Rule_of_thumb
    Improved k-means initial centroids - http://www.ijcsit.com/docs/vol1issue2/ijcsit2010010214.pdf
    """
    # If a threshold is defined, first remove outliers relative to the
    # population before performing the clustering.
    if outlier_threshold:
        outliers = find_outliers(obs, outlier_threshold, False)
        obs = [ob for i, ob in enumerate(obs) if i not in outliers]
    else:
        outliers = []

    # Manually calculate the standard deviation for the matrix in order to
    # de-normalize the resulting centroids. Otherwise, the centroids would
    # be relative to the normalized dimensions rather than the original.
    nobs = numpy.array(obs)
    std = nobs.std(axis=0)
    wobs = nobs / std

    # How many observations?
    n = len(nobs)

    # How many clusters?
    if k is None:
        k = int(math.sqrt(n / 2))

    # Find midpoint of each cluster in a the sorted observations as the
    # initial centroids.
    step = n / k
    offset = step / 2
    initial_centroids = numpy.sort(wobs, axis=0)[offset::step]

    # Perform the clustering then calculate the centroid each observation
    # is associated with and it's distance. Centroids returned here are
    # based on the normalized observations.
    centroids, _ = vq.kmeans(wobs, initial_centroids)
    indexes, distances = vq.vq(wobs, centroids)

    # Re-denormalize centroids for downstream use
    return {
        'centroids': centroids * std,
        'indexes': indexes,
        'distances': distances,
        'outliers': outliers,
    }



import functools
from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import LazyObject
from django.conf import settings as django_settings
from avocado.conf import global_settings


class Settings(object):
    def __init__(self, settings_dict):
        # set the initial settings as defined in the global_settings
        for setting in iter(dir(global_settings)):
            if setting == setting.upper():
                setattr(self, setting, getattr(global_settings, setting))

        # iterate over the user-defined settings and override the default
        # settings
        for key, value in iter(settings_dict.items()):
            if key == key.upper():
                setattr(self, key, value)


class LazySettings(LazyObject):
    def _setup(self):
        self._wrapped = Settings(getattr(django_settings, 'AVOCADO', {}))


settings = LazySettings()


# Keep track of the officially supported apps and libraries used for various
# features.
OPTIONAL_DEPS = {
    'haystack': False,
    'numpy': False,
    'scipy': False,
    'actstream': False,
    'openpyxl': False,
    'guardian': False,
}

# Full-text search engine
try:
    import haystack
    OPTIONAL_DEPS['haystack'] = True
except (ImportError, ImproperlyConfigured):
    pass

# High performance data structures
try:
    import numpy
    OPTIONAL_DEPS['numpy'] = True
except ImportError:
    pass

# Statistical functions
try:
    import scipy
    OPTIONAL_DEPS['scipy'] = True
except ImportError:
    pass

# Activity streams
try:
    import actstream
    if 'actstream' in django_settings.INSTALLED_APPS:
        OPTIONAL_DEPS['acstream'] = True
except ImportError:
    pass

# Native MS Excel reading/writing support
try:
    import openpyxl
    OPTIONAL_DEPS['openpyxl'] = True
except ImportError:
    pass

# Generic object-level permissions
try:
    import guardian
    OPTIONAL_DEPS['guardian'] = True
except ImportError:
    pass


def requires_dep(lib):
    "Decorator for functions that require a supported third-party library."
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if not OPTIONAL_DEPS[lib]:
                raise ImproperlyConfigured('{0} must be installed to use this feature.'.format(lib))
            return f(*args, **kwargs)
        return wrapper
    return decorator



from django.db import transaction
from django.contrib import admin
from django.contrib.admin import SimpleListFilter
from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _
from avocado.models import DataField, DataConcept, DataCategory, DataConceptField
from avocado.forms import DataFieldAdminForm


class PublishedAdmin(admin.ModelAdmin):
    list_per_page = 25
    save_as = True
    actions_on_bottom = True

    # Filter by status
    list_filter = ('published', 'archived')

    # Archived are at the bottom, followed by unpublished, then by name
    ordering = ('archived', '-published', 'name')

    # List here to enable viewing the data
    readonly_fields = ('created', 'modified')
    # TODO change behavior to use Haystack
    search_fields = ('name', 'description', 'keywords')

    # Actions to toggle published and archived status
    actions = ('mark_published', 'mark_unpublished', 'mark_archived',
        'mark_unarchived')

    # Override to make the description consistent with the rest of
    # the actions
    def get_actions(self, request):
        actions = super(PublishedAdmin, self).get_actions(request)
        if 'delete_selected' in actions:
            actions['delete_selected'][0].short_description = 'Delete'
        return actions

    def mark_published(self, request, queryset):
        queryset.update(published=True)
    mark_published.short_description = 'Publish'

    def mark_unpublished(self, request, queryset):
        queryset.update(published=False)
    mark_unpublished.short_description = 'Unpublish'

    def mark_archived(self, request, queryset):
        queryset.update(archived=True)
    mark_archived.short_description = 'Archive'

    def mark_unarchived(self, request, queryset):
        queryset.update(archived=False)
    mark_unarchived.short_description = 'Unarchive'


class LexiconListFilter(SimpleListFilter):
    title = _('lexicon')

    parameter_name = 'lexicon'

    def lookups(self, request, model_admin):
        return (
            ('1', _('Yes')),
            ('0', _('No')),
        )

    def queryset(self, request, queryset):
        value = self.value()

        if value:
            subquery = queryset.only('app_name', 'model_name', 'field_name')
            ids = [x.pk for x in subquery if x.lexicon]
            if value == '1':
                queryset = queryset.filter(id__in=ids)
            else:
                queryset = queryset.exclude(id__in=ids)
            return queryset


class ObjectSetListFilter(SimpleListFilter):
    title = _('objectset')

    parameter_name = 'objectset'

    def lookups(self, request, model_admin):
        return (
            ('1', _('Yes')),
            ('0', _('No')),
        )

    def queryset(self, request, queryset):
        value = self.value()

        if value:
            subquery = queryset.only('app_name', 'model_name', 'field_name')
            ids = [x.pk for x in subquery if x.objectset]
            if value == '1':
                queryset = queryset.filter(id__in=ids)
            else:
                queryset = queryset.exclude(id__in=ids)
            return queryset


class DataFieldAdmin(PublishedAdmin):
    form = DataFieldAdminForm

    list_display = ('name', 'published', 'archived', 'internal',
        'orphan_status', 'model_name', 'enumerable', 'searchable',
        'is_lexicon', 'is_objectset', 'related_dataconcepts')
    list_filter = ('published', 'archived', 'internal', 'model_name',
        'enumerable', 'searchable', LexiconListFilter, ObjectSetListFilter)
    list_editable = ('published', 'archived', 'internal', 'enumerable',
        'searchable')

    search_fields = ('name', 'description', 'keywords')
    readonly_fields = ('created', 'modified', 'data_modified')
    actions = ('mark_published', 'mark_unpublished', 'mark_archived',
        'mark_unarchived', 'create_dataconcept_multi',
        'create_dataconcept_single')

    fieldsets = (
        (None, {
            'fields': (
                'name',
                'name_plural',
                'description',
                'keywords',
                'unit',
                'unit_plural',
                'published',
                'archived',
            ),
        }),

        ('Internal Use', {
            'fields': ('internal',),
            'description': 'Flag as internal if this concept is is intended '\
                'for programmatic access only.',
        }),

        ('Query Modifiers', {
            'fields': ('translator', 'enumerable', 'searchable')
        }),

        ('Data Source', {
            'description': ('These fields should not be altered in-place. '
                'If where this field is located has changed, change these '
                ' values and create a new field by using "Save as new".'),
            'classes': ('collapse',),
            'fields': ('app_name', 'model_name', 'field_name', 'data_source'),
        }),

        ('Times of Interest', {
            'fields': ('created', 'modified', 'data_modified'),
        }),
    )

    def orphan_status(self, obj):
        if obj.model is None:
            return 'Unknown Model'
        if obj.field is None:
            return 'Unknown Model Field'
        return 'OK'
    orphan_status.short_description = 'Orphan Status'

    def is_lexicon(self, obj):
        return obj.lexicon
    is_lexicon.short_description = 'Is Lexicon'

    def is_objectset(self, obj):
        return obj.objectset
    is_objectset.short_description = 'Is ObjectSet'

    def related_dataconcepts(self, obj):
        queryset = obj.concepts.only('id', 'name')
        reverse_name = 'admin:avocado_dataconcept_change'
        urlize = lambda x: '<a href="{0}">{1}</a>'.format(reverse(reverse_name, args=[x.id]), x.name)
        return '<br>'.join(map(urlize, queryset)) or None
    related_dataconcepts.short_description = 'Related Data Concepts'
    related_dataconcepts.allow_tags = True

    def create_dataconcept_multi(self, request, queryset):
        for datafield in queryset:
            DataConcept.objects.create_from_field(datafield)
    create_dataconcept_multi.short_description = 'Create Data Concept for Each'

    @transaction.commit_on_success
    def create_dataconcept_single(self, request, queryset):
        fields = list(queryset)
        max_length = DataConcept._meta.get_field_by_name('name')[0].max_length
        name = ', '.join([f.name for f in fields])[:max_length - 5] + '...'
        concept = DataConcept(name='"{0}"'.format(name))
        concept.save()
        for i, datafield in enumerate(queryset):
            DataConceptField(concept=concept, field=datafield, order=i).save()
    create_dataconcept_single.short_description = 'Create Data Concept for Selected'


class DataConceptFieldInlineAdmin(admin.TabularInline):
    model = DataConceptField


class DataConceptAdmin(PublishedAdmin):
    list_display = ('name', 'published', 'archived', 'internal', 'category',
        'queryview', 'formatter_name', 'sortable', 'related_datafields')
    list_editable = ('published', 'archived', 'internal', 'category',
        'queryview', 'formatter_name', 'sortable')
    inlines = [DataConceptFieldInlineAdmin]

    fieldsets = (
        (None, {
            'fields': (
                'name',
                'name_plural',
                'description',
                'keywords',
                'category',
                'published',
                'archived',
            ),
        }),

        ('Internal Use', {
            'fields': ('internal', 'ident'),
            'description': 'Flag as internal if this concept is is intended '\
                'for programmatic access only.',
        }),

        ('Modifiers', {
            'fields': ('queryview', 'formatter_name', 'sortable'),
        }),

        ('Times of Interest', {
            'fields': ('created', 'modified'),
        }),
    )

    def related_datafields(self, obj):
        queryset = obj.fields.only('id', 'name')
        reverse_name = 'admin:avocado_datafield_change'
        urlize = lambda x: '<a href="{0}">{1}</a>'.format(reverse(reverse_name, args=[x.id]), x.name)
        return '<br>'.join(map(urlize, queryset)) or None
    related_datafields.short_description = 'Related Data Fields'
    related_datafields.allow_tags = True


class DataCategoryAdmin(admin.ModelAdmin):
    model = DataCategory

    list_display = ('name', 'parent', 'order')


admin.site.register(DataField, DataFieldAdmin)
admin.site.register(DataConcept, DataConceptAdmin)
admin.site.register(DataCategory, DataCategoryAdmin)



from optparse import make_option
from django.core.management.base import BaseCommand
from django.db import transaction
from avocado.core import utils
from avocado.models import DataField
from avocado.legacy import models as legacy


class Command(BaseCommand):
    """
    SYNOPSIS::

        python manage.py avocado legacy [options]

    DESCRIPTION:

        Utilities for migrating Avocado 1.x metadata to the
        Avocado 2.x data model.

    OPTIONS:

        ``--force`` - Forces legacy description data to override existing
            fields that have been synced.

        ``--no-input`` - Prevent user prompts and assumes the default behavior
            during migration.
    """

    help = '\n'.join([
        'Utilities for migrating Avocado 1.x metadata to the',
        'Avocado 2.x data model.',
    ])

    option_list = BaseCommand.option_list + (
        make_option('-f', '--force', action='store_true',
            dest='force', default=False,
            help='During a migration, force and update on existing fields'),

        make_option('--no-input', action='store_true',
            dest='no_input', default=False,
            help='Prevents user prompts and performs the default bahavior'),
    )

    @transaction.commit_on_success
    def handle(self, *args, **options):
        force = options.get('force')
        no_input = options.get('no_input')

        total_migrated = 0

        for lf in legacy.Field.objects.iterator():
            try:
                f = DataField.objects.get_by_natural_key(lf.app_name,
                    lf.model_name, lf.field_name)
            except DataField.DoesNotExist:
                f = DataField(app_name=lf.app_name, model_name=lf.model_name,
                    field_name=lf.field_name)

            qualified_name = '({0}) {1}.{2}'.format(f.app_name, f.model_name,
                f.field_name)

            if f.pk and not force:
                print '{0} already exists. Skipping...'.format(qualified_name)
                continue

            # Check if this is an orphan
            if not f.field:
                print '{0} is orphaned. Skipping...'.format(qualified_name)
                continue

            # Map various fields
            f.name = lf.name
            f.description = lf.description
            f.keywords = lf.keywords
            f.translator = lf.translator
            f.group_id = lf.group_id

            print 'Migrating...\t{0}'.format(qualified_name)

            flags = utils.get_heuristic_flags(f)
            f.__dict__.update(flags)

            # Disagreement with enumerable status
            if not no_input and f.enumerable != lf.enable_choices:
                if lf.enable_choices:
                    override = raw_input('"{0}" is marked as enumerable, but '
                        'does not qualify to be enumerable. Override? '
                        '[y/N] '.format(qualified_name))
                else:
                    override = raw_input('"{0}" is not marked as enumerable, '
                        'but qualifies to be enumerable. Override? '
                        '[y/N] '.format(qualified_name))

                if override.lower() == 'y':
                    f.enumerable = lf.enable_choices

            f.save()
            f.sites = lf.sites.all()

            total_migrated += 1

        print 'Fields migrated:\t{0}'.format(total_migrated)



import django
from django.db import models, transaction, IntegrityError


class ObjectSetError(Exception):
    def __init__(self, msg=None):
        if msg is None:
            msg = 'ObjectSet instance needs to have a primary key ' \
                'before set operations can be used.'
        Exception.__init__(self, msg)


class ObjectSet(models.Model):
    """Enables persisting a materialized set of IDs for some object type.
    `ObjectSet` must be subclassed to add the many-to-many relationship
    to the _object_ of interest.
    """
    name = models.CharField(max_length=100, null=True, blank=True)
    description = models.TextField(null=True, blank=True)

    # In cases where an object set has been created as a result of a
    # data context, a reference can be stored for performing updates
    # as more data is added.
    context = models.OneToOneField('avocado.DataContext', null=True,
        blank=True, editable=False)

    # Stored count of the set size. Since most operations are incremental
    # and are applied with objects in memory, this is a more efficient
    # way to keep track of the count as suppose to performing a database
    # count each time.
    count = models.PositiveIntegerField(default=0, editable=False)

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    # Name of field which contains the set of objects
    set_object_rel = None

    class Meta(object):
        abstract = True

    @property
    def _set_object_class(self):
        return getattr(self.__class__, self.set_object_rel).through

    @property
    def _object_field(self):
        opts = self._set_object_class._meta
        return opts.get_field_by_name('set_object')[0]

    @property
    def _object_class(self):
        return self._object_field.rel.to

    @property
    def _all_set_objects(self):
        "Returns a SetObject queryset containing all objects in this set."
        return self._set_object_class.objects.filter(object_set=self)

    @property
    def _set_objects(self):
        "Returns a SetObject queryset containing all non-removed objects."
        return self._all_set_objects.filter(removed=False)

    def __unicode__(self):
        return unicode(self.name)

    def __len__(self):
        "Returns the length (size) of this set."
        return self.count

    def __nonzero__(self):
        return True

    def __iter__(self):
        objects = self._object_class.objects.all()
        pks = self._set_objects.values_list('set_object__pk')
        return iter(objects.filter(pk__in=pks))

    def __contains__(self, obj):
        "Returns True if `obj` is in this set."
        return self._set_objects.filter(set_object=obj).exists()

    def __and__(self, other):
        "Performs an intersection of this set and `other`."
        raise NotImplemented

    def __or__(self, other):
        "Performs an union of this set and `other`."
        raise NotImplemented

    def __xor__(self, other):
        "Performs an exclusive union of this set and `other`."
        raise NotImplemented

    def __sub__(self, other):
        "Removes objects from this set that are in `other`."
        raise NotImplemented

    def __iand__(self, other):
        "Performs an inplace intersection of this set and `other`."
        raise NotImplemented

    def __ior__(self, other):
        "Performs and inplace union of this set and `other`."
        raise NotImplemented

    def __ixor__(self, other):
        "Performs an inplace exclusive union of this set and `other`."
        raise NotImplemented

    def __isub__(self, other):
        "Inplace removal of objects from this set that are in `other`."
        raise NotImplemented

    def _check_pk(self):
        if not self.pk:
            raise ObjectSetError

    def _check_type(self, obj):
        if not isinstance(obj, self._object_class):
            raise TypeError("Only objects of type '{0}' can be added to the "\
                "set".format(self._object_class.__name__))

    def _add(self, obj, added):
        """Check for an existing object that has been removed and mark
        it has not removed, otherwise create a new object and mark it
        as added.
        """
        self._check_type(obj)
        try:
            _obj = self._all_set_objects.get(set_object=obj)
            # Already exists, nothing to do
            if _obj.removed is False:
                return False
            _obj.removed = False
            _obj.added = added
        except self._set_object_class.DoesNotExist:
            _obj = self._set_object_class(set_object=obj, object_set=self,
                added=added)
        _obj.save()
        return True

    @transaction.commit_on_success
    def bulk(self, objs, added=False):
        """Attempts to bulk load objects. Although this is the most efficient
        way to add objects, if any fail to be added, none will be added.

        This should be used when the set is empty and needs to be populated.
        """
        if django.VERSION < (1, 4):
            raise EnvironmentError('This method requires Django 1.4 or above')

        self._check_pk()
        _objs = []
        loaded = 0
        for obj in iter(objs):
            self._check_type(obj)
            _objs.append(self._set_object_class(object_set=self,
                set_object=obj, added=added))
            loaded += 1
        self._set_object_class.objects.bulk_create(_objs)
        self.count += loaded
        self.save()
        return loaded

    @transaction.commit_on_success
    def add(self, obj, added=False):
        "Adds `obj` to the set."
        self._check_pk()
        added = self._add(obj, added)
        if added:
            self.count += 1
            self.save()
        return added

    @transaction.commit_on_success
    def remove(self, obj):
        "Removes `obj` from the set."
        self._check_pk()
        self._check_type(obj)
        removed = self._set_objects.filter(set_object=obj,
            object_set=self).update(removed=True)
        if removed:
            self.count -= 1
            self.save()
        return bool(removed)

    @transaction.commit_on_success
    def update(self, objs, added=True):
        "Update the current set with the objects not already in the set."
        self._check_pk()
        added = 0
        for obj in iter(objs):
            added += int(self._add(obj, added))
        self.count += added
        self.save()
        return added

    @transaction.commit_on_success
    def clear(self):
        "Remove all objects from the set."
        self._check_pk()
        removed = self._set_objects.update(removed=True)
        self.count = 0
        self.save()
        return removed

    @transaction.commit_on_success
    def replace(self, objs):
        "Replace the current set with the new objects."
        self._check_pk()
        self._set_objects.update(removed=True)
        self.count = 0
        return self.update(objs)

    def flush(self):
        "Deletes objects in the set marked as `removed`."
        self._check_pk()
        return self._all_set_objects.filter(removed=True).delete()


class SetObject(models.Model):
    """Adds additional information about the objects that have been `added`
    and `removed` from the original set.

    For instance, additional objects that are added which do not match the
    conditions currently associated with the `ObjectSet` should be flagged
    as `added`. If in the future they match the conditions, the flag can be
    removed.

    Any objects that are removed from the set should be marked as `removed`
    even if they were added at one time. This is too keep track of the objects
    that have been explicitly removed from the set.

    To implement, the `set_object` and `object_set` fields must be defined:

    class BookSetObject(ObjectSet):
        object_set = models.ForeignKey(BookSet)
        set_object = models.ForeignKey(Book)

    """
    added = models.BooleanField(default=False)
    removed = models.BooleanField(default=False)

    class Meta(object):
        abstract = True
        unique_together = ('object_set', 'set_object')



import math
from django.core.paginator import Paginator, Page


class BufferedPaginator(Paginator):
    """A subclass of the Django Paginator class that allows for explicitly
    setting the `_count` attribute i.e. the theoretical size of
    `object_list`. This removes the need for the Paginator to do determine
    the size of the `object_list` which improves performance especially for
    complex queries (and removes the need to actually pass in the
    `objec_list`).

    The `object_list` that is passed in as an argument should represent the
    *buffered* part of the whole `object_list`. This works in conjunction
    with the `BufferedPage` which can test whether or not a particular page
    is actually stored in the `BufferedPaginator` object.

    If `object_list` is not passed in, all calculations will still be
    available, but any operations that act on the `object_list` will
    throw an error, unless an `object_list` is passed in.

    The use of the class is primarily for large data sets in which it is
    impractical to store the entire `object_list` into a paginator.
    Overriding `_count` allows for the methods to still be used.

    `buf_size` represents the size of the buffer, i.e. number of rows that will
    be available at any given time and in most cases the size of `object_list`
    assuming `_count` is greater than `buf_size`..
    """
    def __init__(self, count, object_list=None, offset=0, buf_size=None, *args, **kwargs):
        if offset > count:
            raise ValueError('"offset" cannot be greater than the "count"')

        super(BufferedPaginator, self).__init__(object_list, *args, **kwargs)

        self._count = count

        # Negative offsets are relative to to `count`
        if offset < 0:
            offset = count + offset

        self.offset = offset
        self.buf_size = buf_size or count

    def page(self, number):
        """Returns a `BufferedPage` object representing the slice of data
        relative to the `per_page` number.
        """
        number = self.validate_number(number)

        if self.object_list is not None:
            bottom = (number - 1) * self.per_page - self.offset

            if bottom < 0:
                bottom = 0

            top = bottom + self.per_page

            if top + self.orphans >= self.count:
                top = self.count

            object_list = self.object_list[bottom:top]
        else:
            object_list = None

        return BufferedPage(number, self, object_list)

    def cached_page_indices(self):
        "Returns the indices of the first and last pages that are cached."
        # The number of rows per page is greater than we will ever have in
        # the buffer at any given time. We can only provide partial results
        # The beginning items are missing.. no good
        if self.buf_size < self.per_page and self.offset > 0:
            return (0, 0)

        pages_of_offset = math.ceil(self.offset / float(self.per_page))

        first = int(pages_of_offset) + 1
        last = first + math.ceil(self.buf_size / float(self.per_page))

        return (int(first), int(last))

    def cached_pages(self):
        "Returns the number of cached pages."
        first, last = self.cached_page_indices()
        return last - first

    def get_overlap(self, offset, buf_size=None):
        """Calculates the overlap of the buffered rows relative to a new offset
        and (optionally) buf_size. There are two sets of outputs here including
        the start and end terminals::

            (overlaps, (start_offset, start_limit), (end_offset, end_limit))

        The offsets and limits can be used for list slicing or for SQL OFFSET/
        LIMIT clauses.
        """
        buf_size = buf_size or self.buf_size

        overlap = True
        start_offset = start_limit = end_offset = end_limit = None

        # If the start terminal is within the current range, then there will be
        # a slice of data to be appended
        if self.offset <= offset <= (self.offset + self.buf_size):
            end_offset = (self.offset + self.buf_size) + 1
            end_limit = buf_size - (end_offset - offset) + 1

        # If the end terminal is within the current range, then there will be a
        # slice of data to be prepended
        elif self.offset <= (offset + buf_size) <= (self.offset + self.buf_size):
            start_offset = offset
            start_limit = self.offset - start_offset

        # Check to see if the current range is within the start and end
        # terminals
        elif offset <= self.offset and (self.offset + self.buf_size) <= (offset + buf_size):
            start_offset = offset
            start_limit = self.offset - start_offset
            end_offset = (self.offset + self.buf_size) + 1
            end_limit = (offset + buf_size) - end_offset + 1

        # No overlap at all
        else:
            overlap = False
            start_offset = offset
            start_limit = buf_size

        return (overlap, (start_offset, start_limit), (end_offset, end_limit))


class BufferedPage(Page):
    """A subclass of the Django `Page` class which adds an additional method to
    determine if this page is in cache. The determination is with respect to
    the `paginator`, `offset` and `buf_size' attributes.
    """
    def __init__(self, number, paginator, object_list=None):
        self.number = number
        self.paginator = paginator
        self.object_list = object_list

    def in_cache(self):
        "Returns true if this page is in the cache (buffer)."
        first, last = self.paginator.cached_page_indices()
        return first <= self.number < last

    def offset(self):
        "Returns a zero-based offset of this page."
        return max(self.start_index(), 1) - 1

    def get_list(self, object_list=None):
        """Get the list of objects for this page.

        If an `object_list` is passed in, the slice corresponding to this
        page is returned.
        """
        if object_list is not None:
            start = self.start_index() - self.paginator.offset - 1
            end = self.end_index() - self.paginator.offset
            return object_list[start:end]

        if self.object_list is not None:
            return self.object_list

        return None

    def page_links(self, before_after_count=2, first_last_count=0):
        pages = []
        # only process if the current page number is greater than the number of
        # required links.
        if self.number > (first_last_count + before_after_count + 1):
            # append first pages, e.g. [1, 2] if `first_last_count' is 2
            for i in range(1, first_last_count + 1):
                pages.append(i)

            pages.append(None)
            # append the leading pages relative to the current page, e.g.
            # [10, 11] if `before_after_count' is 2 and current page is 12
            for i in range(self.number - before_after_count, self.number):
                pages.append(i)

        else:
            for i in range(1, self.number):
                pages.append(i)

        # if there is still a gap until `self.paginator.num_pages' is reached, add trailing pages
        if (self.number + first_last_count + before_after_count) < self.paginator.num_pages:
            for i in range(self.number, self.number + before_after_count + 1):
                pages.append(i)

            pages.append(None)
            for i in range(self.paginator.num_pages + 1 - first_last_count, self.paginator.num_pages + 1):
                pages.append(i)

        else:
            for i in range(self.number, self.paginator.num_pages + 1):
                pages.append(i)

        return pages



import logging
from collections import defaultdict
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from django.utils.encoding import force_unicode
from avocado.core import loader

log = logging.getLogger(__name__)


class FormatterException(Exception):
    pass


def unique_keys(fields):
    """Takes a list of fields and generated a unique list of keys based
    based on the field's natural_key.
    """
    def model_prefix(field, key):
        return '{0}__{1}'.format(field.model_name, key)

    def app_prefix(field, key):
        return '{0}__{1}'.format(field.app_name, model_prefix(field, key))

    # Evaluate in case this is QuerySet..
    fields = list(fields)

    keys = []
    names = []
    models = defaultdict(int)

    # Starting set of field names
    for f in fields:
        names.append(f.field_name)
        models[f.model] += 1

    # For consistency, any keys that do conflict, all occurences will
    # be prefixed the same way either with the 'app' or 'model' level
    prefixed = {}

    for i, name in enumerate(names):
        field = fields[i]

        # Check is the prefix is pending
        if name in prefixed:
            key = prefixed[name](field, name)
        elif names.count(name) == 1:
            key = name
        else:
            if models[field.model] == 1:
                key = model_prefix(field, name)
                method = model_prefix
            else:
                key = app_prefix(field, name)
                method = app_prefix

            # Mark this field name to prefixed from now on
            prefixed[name] = method
        keys.append(key)

    return keys


class Formatter(object):
    """Provides support for the core data formats with sensible defaults
    for handling converting Python datatypes to their formatted equivalent.

    Each core format method must return one of the following:
        Single formatted Value
        OrderedDict/sequence of key-value pairs

        If the format method is unable to do either of these for the given
        value a FormatterException must be raised.

        ``values`` - A list, tuple or OrderedDict containing the values to
        be formatted. If a list or tuple is passed, it will be wrapped in
        an OrderedDict for keyword access in the format methods.

        ::

            values = ['Bob', 'Smith']

    """
    default_formats = ('boolean', 'number', 'string')

    def __init__(self, concept=None, keys=None):
        "Passing in a concept takes precedence over `keys`."
        if not keys and not concept:
            raise ValueError('A concept or list of keys must be supplied.')

        self.concept = concept
        self.fields = None

        if concept:
            fields = list(concept.fields.order_by('concept_fields__order'))
            self.keys = unique_keys(fields)
            self.fields = OrderedDict(zip(self.keys, fields))
        else:
            self.keys = keys

    def __call__(self, values, preferred_formats=None, **context):
        # Create a copy of the preferred formats since each set values may
        # be processed slightly differently (e.g. mixed data type in column)
        # which could cause exceptions that would not be present during
        # processing of other values
        if preferred_formats is None:
            preferred_formats = self.default_formats
        preferred_formats = list(preferred_formats) + ['raw']

        # Create a OrderedDict of the values relative to the
        # concept fields objects the values represent. This
        # enables key-based access to the values rather than
        # relying on position.
        if not isinstance(values, OrderedDict):
            # Wrap single values
            if not isinstance(values, (list, tuple)):
                values = [values]
            values = OrderedDict(zip(self.keys, values))

        # Iterate over all preferred formats and attempt to process the values.
        # For formatter methods that process all values must be tracked and
        # attempted only once. They are removed from the list once attempted.
        # If no preferred multi-value methods succeed, each value is processed
        # independently with the remaining formats
        for f in iter(preferred_formats):
            method = getattr(self, 'to_{0}'.format(f), None)
            # This formatter does not support this format, remove it
            # from the available list
            if not method:
                preferred_formats.pop(0)
                continue

            # The implicit behavior when handling multiple values is to process
            # them independently since, in most cases, they are not dependent
            # on one another, but rather should be represented together since
            # the data is related. A formatter method can be flagged to process
            # all values together by setting the attribute
            # `process_multiple=True`. we must # check to if that flag has been
            # set and simply pass through the values and context to the method
            # as is. if ``process_multiple`` is not set, each value is handled
            # independently
            if getattr(method, 'process_multiple', False):
                try:
                    output = method(values, fields=self.fields,
                        concept=self.concept, process_multiple=True, **context)
                    if not isinstance(output, dict):
                        return OrderedDict([(self.concept.name, output)])
                    return output
                # Remove from the preferred formats list since it failed
                except Exception:
                    log.exception('Multi-value formatter error')
                    preferred_formats.pop(0)

        # The output is independent of the input. Formatters may output more
        # or less values than what was entered.
        output = OrderedDict()

        # Attempt to process each
        for i, (key, value) in enumerate(values.iteritems()):
            for f in preferred_formats:
                method = getattr(self, 'to_{0}'.format(f))
                try:
                    field = self.fields[key] if self.fields else None
                    fvalue = method(value, field=field, concept=self.concept,
                        process_multiple=False, **context)
                    if isinstance(fvalue, dict):
                        output.update(fvalue)
                    else:
                        output[key] = fvalue
                    break
                except Exception:
                    log.exception('Single-value formatter error')
        return output

    def __contains__(self, choice):
        return hasattr(self, 'to_{0}'.format(choice))

    def __unicode__(self):
        return u'{0}'.format(self.name or self.__class__.__name__)

    def to_string(self, value, **context):
        # Attempt to coerce non-strings to strings. Depending on the data
        # types that are being passed into this, this may not be good
        # enough for certain datatypes or complext data structures
        if value is None:
            return u''
        return force_unicode(value, strings_only=False)

    def to_boolean(self, value, **context):
        # If value is native True or False value, return it
        if type(value) is bool:
            return value
        raise FormatterException('Cannot convert {0} to boolean'.format(value))

    def to_number(self, value, **context):
        # Attempts to convert a number. Starting with ints and floats
        # Eventually create to_decimal using the decimal library.
        if type(value) is int or type(value) is float:
            return value
        if isinstance(value, basestring):
            try:
                value = int(value)
            except (ValueError, TypeError):
                value = float(value)
            return value
        raise FormatterException('Cannot convert {0} to number'.format(value))

    def to_coded(self, value, **context):
        # Attempts to convert value to its coded representation
        if 'field' in context:
            for key, coded in context['field'].coded_values:
                if key == value:
                    return coded
        raise FormatterException('No coded value for {0}'.format(value))

    def to_raw(self, value, **context):
        return value


class RawFormatter(Formatter):
    def __call__(self, values, *args, **kwargs):
        preferred_formats = ['raw']
        return super(RawFormatter, self).__call__(values, preferred_formats)


registry = loader.Registry(default=Formatter, register_instance=False)
loader.autodiscover('formatters')



import inspect
from django.conf import settings
from django.utils.importlib import import_module
from django.core.exceptions import ImproperlyConfigured


class AlreadyRegistered(Exception):
    pass


class NotRegistered(Exception):
    pass


class Registry(object):
    "Simple class that keeps track of a set of registered classes."
    def __init__(self, default=None, register_instance=True):
        if register_instance and inspect.isclass(default):
            default = default()
        self.register_instance = register_instance
        self.default = default
        self._registry = {}
        if default:
            self.register(default)

    def __getitem__(self, name):
        return self._registry.get(name, self.default)

    def get(self, name):
        return self.__getitem__(name)

    def register(self, obj, name=None):
        """Registers a class with an optional name. The class name will be used
        if not supplied.
        """
        if inspect.isclass(obj):
            name = name or obj.__name__
            # Create an instance if instances should be registered
            if self.register_instance:
                obj = obj()
        else:
            name = name or obj.__class__.__name__

        if name in self._registry:
            raise AlreadyRegistered('The class {0} is already registered'.format(name))

        # Check to see if this class should be used as the default for this
        # registry
        if getattr(obj, 'default', False):
            # ensure the default if already overriden is not being overriden
            # again.
            if self.default:
                if self.register_instance:
                    name = self.default.__class__.__name__
                else:
                    name = self.default.__name__
                objtype = 'class' if self.register_instance else 'instance'
                raise ImproperlyConfigured('The default {0} cannot be set '
                    'more than once for this registry ({1} is the default).'.format(objtype, name))

            self.default = obj
        else:
            if name in self._registry:
                raise AlreadyRegistered('Another class is registered with the '
                    'name "{0}"'.format(name))

            self._registry[name] = obj

    def unregister(self, name):
        """Unregisters a class. Note that these calls must be made in
        INSTALLED_APPS listed after the apps that already registered the class.
        """
        # Use the name of the class if passed in. Second condition checks for an
        # instance of the class.
        if inspect.isclass(name):
            name = name.__name__
        elif hasattr(name, '__class__'):
            name = name.__class__.__name__
        if name not in self._registry:
            objtype = 'class' if self.register_instance else 'instance'
            raise NotRegistered('No {0} is registered under the name "{1}"'.format(objtype, name))
        self._registry.pop(name)

    @property
    def choices(self):
        "Returns a 2-tuple list of all registered class instance names."
        return sorted((x, x) for x in self._registry.iterkeys())


def autodiscover(module_name):
    """Simple auto-discover for looking through each INSTALLED_APPS for each
    ``module_name`` and fail silently when not found. This should be used for
    modules that have 'registration' like behavior.
    """
    for app in settings.INSTALLED_APPS:
        # Attempt to import the app's ``module_name``.
        try:
            import_module('{0}.{1}'.format(app, module_name))
        except:
            pass



import sys
from setuptools import setup, find_packages
from setuptools.command import install as _install


install_requires = [
    'django>=1.4,<1.5',
    'modeltree>=1.1.2',
    'South==0.7.6',
    # Uses a dependency link below
    'jsonfield>=1.0b',
]

if sys.version_info < (2, 7):
    install_requires.append('ordereddict>=1.1')


class install(_install.install):
    def run(self):
        _install.install.run(self)
        sys.stdout.write('NOTE: For `clustering` support, NumPy must '\
            'installed first, followed by SciPy.\n')

kwargs = {
    'cmdclass': {'install': install},

    # Packages
    'packages': find_packages(exclude=['tests', '*.tests', '*.tests.*', 'tests.*']),
    'include_package_data': True,

    # Dependencies
    'install_requires': install_requires,

    'test_suite': 'test_suite',

    # Test dependencies
    'tests_require': [
        'django-guardian',
        'django-haystack',
        'whoosh',
        'openpyxl',
        'scipy',
        'numpy',
        'coverage',
        'python-memcached',
    ],

    # Optional dependencies
    'extras_require': {
        # Granular permission
        'permissions': ['django-guardian'],
        # Search
        'search': ['django-haystack'],
        # Clustering components.. unforunately SciPy must be installed
        # separately since NumPy is a dependency
        'clustering': ['numpy'],
        # Includes extra exporter dependencies
        'extras': ['openpyxl'],
    },

    'dependency_links': [
        'https://github.com/cbmi/django-jsonfield/zipball/601872f#egg=jsonfield-1.0b',
    ],

    # Metadata
    'name': 'avocado',
    'version': __import__('avocado').get_version(),
    'author': 'Byron Ruth',
    'author_email': 'b@devel.io',
    'description': 'Metadata APIs for Django',
    'license': 'BSD',
    'keywords': 'query metadata harvest orm',
    'url': 'http://cbmi.github.com/avocado/',
    'classifiers': [
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Framework :: Django',
        'Topic :: Internet :: WWW/HTTP',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Information Technology',
    ],
}

setup(**kwargs)



from avocado.core import loader
from avocado.conf import OPTIONAL_DEPS
from _csv import CSVExporter
from _sas import SASExporter
from _r import RExporter
from _json import JSONExporter
from _html import HTMLExporter

registry = loader.Registry(register_instance=False)

registry.register(CSVExporter, 'csv')
registry.register(SASExporter, 'sas')
registry.register(RExporter, 'r')
registry.register(JSONExporter, 'json')
registry.register(HTMLExporter, 'html')

if OPTIONAL_DEPS['openpyxl']:
    from _excel import ExcelExporter
    registry.register(ExcelExporter, 'excel')

loader.autodiscover('exporters')



import inspect
from django.core.serializers.json import DjangoJSONEncoder
from _base import BaseExporter


class JSONGeneratorEncoder(DjangoJSONEncoder):
    "Handle generator objects and expressions."
    def default(self, obj):
        if inspect.isgenerator(obj):
            return list(obj)
        return super(JSONGeneratorEncoder, self).default(obj)


class JSONExporter(BaseExporter):
    file_extension = 'json'
    content_type = 'application/json'
    preferred_formats = ('number', 'string')

    def write(self, iterable, buff=None):
        buff = self.get_file_obj(buff)

        encoder = JSONGeneratorEncoder()
        for chunk in encoder.iterencode(self.read(iterable)):
            buff.write(chunk)
        return buff
